# ************************************************************************************
# ***                C O N F I D E N T I A L  --  H I - S O U R C I N G          ***
# ************************************************************************************
# *                                                                                  *
# *                Project Name : Hi-Sourcing Utilities                              *
# *                                                                                  *
# *                File Name : hi_sourcing.py                                        *
# *                                                                                  *
# *                Programmer : Jinwen Liu                                           *
# *                                                                                  *
# *                Start Date : January 9, 2025                                      *
# *                                                                                  *
# *                Last Update : January 9, 2025                                     *
# *                                                                                  *
# *                Version : 0.0.5                                                   *
# *                                                                                  *
# *-------------------------------------------------------------------------------*
# * Class Description:                                                              *
# *   HiSourcing -- Main class for performing similarity-based data sourcing       *
# *   Uses FAISS (Facebook AI Similarity Search) to find similar entries in a      *
# *   database based on a query dataset. Supports filtering by labels and custom   *
# *   column removal.                                                              *
# *                                                                                *
# * Functions:                                                                     *
# *   Public:                                                                      *
# *   - __init__ -- Initialize with query/database DataFrames and label column    *
# *   - remove_columns_from_df -- Removes specified columns from DataFrames       *
# *   - set_fillna_method -- Sets method for handling NaN values                  *
# *   - fillna -- Fills NaN values using specified method                         *
# *   - run -- Performs similarity search using FAISS after filtering by label    *
# *   - sourcing -- Returns the sourcing results as a DataFrame                   *
# *   - validate -- Compares label counts before and after sourcing              *
# *   Private:                                                                     *
# *   - _get_credentials -- Internal method for credential verification           *
# *       - Prompts user for credentials and returns the input credential        *
# *       - Used for access control in the run method                            *
# *                                                                                *
# * Parameters:                                                                    *
# *   query_df -- DataFrame containing the query data, which is features with labels                                *
# *   db_df -- DataFrame containing the database to search within                 *
# *   label -- Column name for label-based filtering                              *
# *   remove_columns -- List of columns to remove before similarity search        *
# *   fillna_method -- Method to handle missing values                            *
# *   k -- Number of nearest neighbors to retrieve                                *
# *   credentials -- Optional credentials for access control                       *
# *                                                                                *
# *******************************************************************************/

import pandas as pd
import numpy as np
import faiss
from getpass import getpass


class HiSourcing:
    def __init__(self,
                 query_df: pd.DataFrame,
                 db_df: pd.DataFrame,
                 label: str = None,
                 remove_columns: list[str] = None,
                 fillna_method: str = 'zero',
                 k: int = 1000,
                 sourcing_rate: float = 0.0, 
                 credentials: str = None
    ):
        self.query_df_with_label = query_df # assume it has label
        self.db_df_with_label = db_df
        
        self.label = label
        self.remove_columns = remove_columns

        self.fillna_method = fillna_method
        self.k = 1000
        self.sourcing_rate = sourcing_rate
        self.D = None
        self.I = None
        self.indices = []  # Initialize as empty list
        self.return_df = pd.DataFrame()  # Initialize as empty DataFrame
        self.df_after_sourcing = pd.DataFrame()  # Initialize as empty DataFrame
        self.credentials = credentials

        self.df_processing()

    def df_processing(self):
        # if no label is provided
        # it means label has already been removed from dataframe
        # and it is the same for query and db, and READY for indexing
        if self.label is None:
            # query dataframe has to be ready
            self.query_df_no_label = self.query_df_with_label
            self.db_df_no_label = self.db_df_with_label
        
        # otherwise is there is label for query, we need to filter out the label
        # to CREATE query df
        # if there is label for db, we are able to validate
        else:
            # query need to be filtered by label=1
            self.query_df_no_label = self.query_df_with_label[self.query_df_with_label[self.label]==1]

            # no matter database dataframe has label or not
            # everything stays
            self.db_df_no_label = self.db_df_with_label
            
            self.query_df_no_label = self.query_df_no_label.drop(self.label, axis=1)
            self.db_df_no_label = self.db_df_no_label.drop(self.label, axis=1)

        if self.remove_columns:
            self.query_df_no_label = self.remove_columns_from_df(self.query_df_no_label)
            self.db_df_no_label = self.remove_columns_from_df(self.db_df_no_label)
        
        # fillna
        self.query_df_no_label = self.fillna(self.query_df_no_label)
        self.db_df_no_label = self.fillna(self.db_df_no_label)

    def remove_columns_from_df(self, df):
        if not self.remove_columns:
            return df
        return df.drop(self.remove_columns, axis=1, errors='ignore')

    def set_fillna_method(self, method):
        self.fillna_method = method



    def fillna(self, df):
        try:
            if self.fillna_method == 'zero':
                return df.fillna(0)
            elif self.fillna_method == 'mean':
                mean_values = df.mean()
                if mean_values.isna().any():
                    print("Warning: Some columns have all NA values, using 0 for those columns")
                    df = df.fillna(mean_values)
                    return df.fillna(0)  # Fill remaining NAs with 0
                return df.fillna(mean_values)
            elif self.fillna_method == 'median':
                median_values = df.median()
                if median_values.isna().any():
                    print("Warning: Some columns have all NA values, using 0 for those columns")
                    df = df.fillna(median_values)
                    return df.fillna(0)
                return df.fillna(median_values)
            elif self.fillna_method == 'mode':
                mode_values = df.mode().iloc[0]
                if mode_values.isna().any():
                    print("Warning: Some columns have all NA values, using 0 for those columns")
                    df = df.fillna(mode_values)
                    return df.fillna(0)
                return df.fillna(mode_values)
            elif self.fillna_method == 'max':
                max_values = df.max()
                if max_values.isna().any():
                    print("Warning: Some columns have all NA values, using 0 for those columns")
                    df = df.fillna(max_values)
                    return df.fillna(0)
                return df.fillna(max_values)
            elif self.fillna_method == 'min':
                min_values = df.min()
                if min_values.isna().any():
                    print("Warning: Some columns have all NA values, using 0 for those columns")
                    df = df.fillna(min_values)
                    return df.fillna(0)
                return df.fillna(min_values)
            else:
                print("Invalid fillna_method. Using 'zero' method instead.")
                return df.fillna(0)
        except Exception as e:
            print(f"Error in fillna: {str(e)}. Using 'zero' method instead.")
            return df.fillna(0)
    
    def _get_credentials(self):
        """
        Internal method for credential verification.

        Prompts user for credentials and returns the input credential.
        Used for access control in the run method.
        """
        credential = getpass("Please enter your credentials: ")
        return credential

    def indexing(self):
        # credential box
        # input_credential = self._get_credentials()
        # if input_credential != 'hijinwen':
        #     print("Access denied: Invalid credentials")
        #     return False
            
        try:
            
            # Convert to numpy arrays and ensure they're float32
            query_array = self.query_df_no_label.values.astype('float32')
            db_array = self.db_df_no_label.values.astype('float32')
            
            # faiss
            index = faiss.IndexFlatL2(query_array.shape[1])
            index.add(db_array)
            
            # Ensure k doesn't exceed database size
            effective_k = min(self.k, len(self.db_df_no_label))
            self.D, self.I = index.search(query_array, effective_k)

            self.indices = list(set([index for sublist in self.I for index in sublist]))
            
            return self.indices
            
        except Exception as e:
            print(f"Error running sourcing: {str(e)}")
            return False
        
    def sourcing(self):
        """Returns the sourcing results as a DataFrame."""
        self.indexing()
        self.return_df = self.db_df_no_label.iloc[self.indices]
        # self.return_df = self.return_df.drop_duplicates()
        return self.return_df

    def validate(self):
        """Validates the sourcing results by comparing label counts."""
        try:
            if self.indices is None or not self.indices:
                print("No results to validate. Run the sourcing process first.")
                return
                
            # labels in db_db
            self.label_before_sourcing = self.db_df_with_label[self.db_df_with_label[self.label]==1].shape[0]
            
            # return prediction df has no label
            self.df_after_sourcing = self.db_df_with_label.iloc[self.indices]
            # self.df_after_sourcing = self.df_after_sourcing.drop_duplicates()
            self.label_after_sourcing = self.df_after_sourcing[self.df_after_sourcing[self.label]==1].shape[0] 

            print("Label before sourcing: " + str(self.label_before_sourcing))
            print("Label after sourcing: " + str(self.label_after_sourcing))
            print("Number of rows before sourcing: " + str(self.db_df_with_label.shape[0]))
            print("Number of rows after sourcing: " + str(self.df_after_sourcing.shape[0]))

        except Exception as e:
            print(f"Error validating: {str(e)}")
